from ninja import NinjaAPI, Schema

from allianceauth.services.hooks import get_extension_logger

from incursions.models.waitlist import (
    ApprovedFitting, WaitlistCategory, WaitlistCategoryRule,
)


class DNAFittingSchema(Schema):
    name: str
    dna: str
    tier: str
    implant_set: str


class FittingNoteSchema(Schema):
    name: str
    description: str


class FittingResponse(Schema):
    fittingdata: list[DNAFittingSchema] | None
    notes: list[FittingNoteSchema] | None
    logi_rules: list[int] | None  # Im hoping to deprecate this entirely, its dumb


logger = get_extension_logger(__name__)
api = NinjaAPI()


def setup(api: NinjaAPI) -> None:
    FittingsAPIEndpoints(api)


class FittingsAPIEndpoints:

    tags = ["Fittings"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/fittings", response={200: FittingResponse, 403: dict, 404: dict}, tags=self.tags)
        def fittings(request):
            if not request.user.has_perm("incursions.basic_waitlist"):
                logger.warning(f"User {request.user} denied access to fittings")
                return 403, {"error": "Permission denied"}

            fittings = list(ApprovedFitting.objects.select_related("implants", "ship"))

            logi_category, created = WaitlistCategory.objects.get_or_create(name="LOGI")
            if created:
                logger.info("LOGI category was created automatically")

            logi_rules = list(WaitlistCategoryRule.objects.filter(waitlist_category_id=logi_category.pk).values_list("ship__type_id", flat=True))

            logger.info(f"User {request.user} fetched {len(fittings)} fittings and {len(logi_rules)} logi rules")
            return 200, FittingResponse(
                fittingdata=fittings,
                notes=fittings,
                logi_rules=logi_rules
            )



from datetime import datetime, timezone

from ninja import NinjaAPI, Schema

from django.db import transaction
from django.shortcuts import get_object_or_404

from allianceauth.eveonline.models import EveCharacter
from allianceauth.framework.api.user import get_main_character_from_user
from allianceauth.services.hooks import get_extension_logger

from incursions.api.schema import CharacterSchema
from incursions.models.waitlist import CharacterRoles, Role


class RoleSchema(Schema):
    name: str
    description: str


class CommanderSchema(Schema):
    character: CharacterSchema
    role: RoleSchema
    granted_by: CharacterSchema | None
    granted_at: datetime


class CommanderFiltersSchema(Schema):
    name: str
    member_count: int


class CommanderListSchema(Schema):
    commanders: list[CommanderSchema]
    filters: list[CommanderFiltersSchema]


class RequestPayload(Schema):
    character_id: int
    role: str


logger = get_extension_logger(__name__)
api = NinjaAPI()


def setup(api: NinjaAPI) -> None:
    CommandersAPIEndpoints(api)


class CommandersAPIEndpoints:

    tags = ["Commanders"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/commanders", response={200: CommanderListSchema, 403: dict}, tags=self.tags)
        def list_commanders(request):
            if not request.user.has_perm("incursions.view_commanders"):
                logger.warning(f"User {request.user} denied access to commander list")
                return 403, {"error": "Permission denied"}

            commanders = CharacterRoles.objects.select_related("character", "role", "granted_by")
            filters = Role.objects.only("pk", "name")
            logger.info(f"User {request.user} fetched {commanders.count()} commanders and {filters.count()} filters")
            return 200, CommanderListSchema(commanders=list(commanders), filters=list(filters))

        @api.post("/commanders", response={200: dict, 400: dict, 403: dict, 404: dict}, tags=self.tags)
        def assign_commander(request, payload: RequestPayload):
            if not request.user.has_perm("incursions.manage_commanders"):
                logger.warning(f"User {request.user} denied assigning commander role")
                return 403, {"error": "Permission denied"}

            if CharacterRoles.objects.filter(character__character_id=payload.character_id).exists():
                logger.warning(f"Character {payload.character_id} already has a commander role")
                return 400, {"error": "Character already has a role"}

            role = get_object_or_404(Role, name=payload.role)
            character = get_object_or_404(EveCharacter, character_id=payload.character_id)
            issuer = get_main_character_from_user(request.user)

            if role.power > CharacterRoles.objects.get(character=issuer).role.power:
                logger.warning(f"User {request.user} attempted to assign higher power role")
                return 403, {"error": "Permission denied, You don't have the power to assign this role"}

            with transaction.atomic():
                commander = CharacterRoles.objects.create(
                    character=character,
                    role=role,
                    granted_by=issuer,
                    granted_at=datetime.now(timezone.utc)
                )

            logger.info(f"Assigned commander role {role.name} to character {character.pk} by user {request.user}")
            return 200, {"status": "Commander assigned", "id": commander.pk}

        @api.get("/commanders/roles", response={200: list[str], 403: dict}, tags=self.tags)
        def assignable_roles(request):
            if not request.user.has_perm("incursions.view_commanders"):
                logger.warning(f"User {request.user} denied access to roles list")
                return 403, {"error": "Permission denied"}

            roles = list(Role.objects.only("name").values_list("name", flat=True))
            logger.info(f"User {request.user} fetched {len(roles)} assignable roles")
            return 200, roles

        @api.get("/commanders/{character_id}", response={200: str, 403: dict}, tags=self.tags)
        def lookup_commander(request, character_id: int):
            if not request.user.has_perm("incursions.view_commanders"):
                logger.warning(f"User {request.user} denied lookup for character {character_id}")
                return 403, {"error": "Permission denied"}

            character_role = get_object_or_404(CharacterRoles.objects.select_related("role"), character__character_id=character_id)
            logger.info(f"User {request.user} looked up role {character_role.role.name} for character {character_id}")
            return 200, character_role.role.name

        @api.delete("/commanders/{character_id}", response={200: dict, 403: dict, 404: dict}, tags=self.tags)
        def revoke_commander(request, character_id: int):
            if not request.user.has_perm("incursions.manage_commanders"):
                logger.warning(f"User {request.user} denied revoke for character {character_id}")
                return 403, {"error": "Permission denied"}

            try:
                commander = CharacterRoles.objects.select_related("role").get(character__character_id=character_id)
            except CharacterRoles.DoesNotExist:
                logger.error(f"Commander role not found for character {character_id}")
                return 404, {"error": "Commander role not found"}

            issuer = get_main_character_from_user(request.user)
            if commander.role.power > CharacterRoles.objects.get(character=issuer).role.power:
                logger.warning(f"User {request.user} lacks permission to revoke role from character {character_id}")
                return 403, {"error": "Permission denied, You don't have the power to revoke this role"}

            deleted, _ = CharacterRoles.objects.filter(character__character_id=character_id).delete()
            logger.info(f"Commander role revoked for character {character_id} by user {request.user}")
            return 200, {"status": "Commander role revoked"}



from collections import defaultdict
from datetime import timedelta

from ninja import NinjaAPI, Schema

from django.db.models import Count, DurationField, ExpressionWrapper, F, Sum
from django.db.models.functions import TruncMonth
from django.utils import timezone

from incursions.models.waitlist import FleetActivity, WaitlistEntryFit


def get_fleet_seconds_by_hull_by_month():
    result = defaultdict(lambda: defaultdict(float))
    qs = (
        FleetActivity.objects
        .annotate(month=TruncMonth('first_seen'))
        .annotate(duration=ExpressionWrapper(F('last_seen') - F('first_seen'), output_field=DurationField()))
        .values('month', 'ship__name')
        .annotate(total=Sum('duration'))
    )
    for row in qs:
        month = row['month'].strftime("%Y-%m")
        result[month][row['ship__name']] += row['total'].total_seconds()
    return result


def get_xes_by_hull_by_month():
    result = defaultdict(lambda: defaultdict(float))
    qs = (
        WaitlistEntryFit.objects
        .annotate(month=TruncMonth('waitlist_entry__joined_at'))
        .values('month', 'fit__ship__name')
        .annotate(count=Count('id'))
    )
    for row in qs:
        month = row['month'].strftime("%Y-%m")
        result[month][row['fit__ship__name']] += row['count']
    return result


def get_fleet_seconds_by_month():
    result = defaultdict(float)
    qs = (
        FleetActivity.objects
        .annotate(month=TruncMonth('first_seen'))
        .annotate(duration=ExpressionWrapper(F('last_seen') - F('first_seen'), output_field=DurationField()))
        .values('month')
        .annotate(total=Sum('duration'))
    )
    for row in qs:
        month = row['month'].strftime("%Y-%m")
        result[month] += row['total'].total_seconds()
    return result


def get_pilots_by_month():
    result = defaultdict(int)
    qs = (
        FleetActivity.objects
        .annotate(month=TruncMonth('first_seen'))
        .values('month')
        .annotate(pilots=Count('character', distinct=True))
    )
    for row in qs:
        month = row['month'].strftime("%Y-%m")
        result[month] = row['pilots']
    return result


def get_xes_by_hull_28d():
    result = defaultdict(float)
    since = timezone.now() - timedelta(days=28)
    qs = (
        WaitlistEntryFit.objects
        .filter(waitlist_entry__joined_at__gte=since)
        .values('fit__ship__name')
        .annotate(count=Count('id'))
    )
    for row in qs:
        result[row['fit__ship__name']] += row['count']
    return result


def get_fleet_seconds_by_hull_28d():
    result = defaultdict(float)
    since = timezone.now() - timedelta(days=28)
    qs = (
        FleetActivity.objects
        .filter(first_seen__gte=since)
        .annotate(duration=ExpressionWrapper(F('last_seen') - F('first_seen'), output_field=DurationField()))
        .values('ship__name')
        .annotate(total=Sum('duration'))
    )
    for row in qs:
        result[row['ship__name']] += row['total'].total_seconds()
    return result


def get_x_vs_time_by_hull_28d():
    xes = get_xes_by_hull_28d()
    seconds = get_fleet_seconds_by_hull_28d()
    result = defaultdict(dict)
    for hull in xes:
        result[hull]['xes'] = xes[hull]
        result[hull]['seconds'] = seconds.get(hull, 0)
    return result


def get_time_spent_in_fleet_by_month():
    result = defaultdict(lambda: defaultdict(float))
    qs = (
        FleetActivity.objects
        .annotate(month=TruncMonth('first_seen'))
        .annotate(duration=ExpressionWrapper(F('last_seen') - F('first_seen'), output_field=DurationField()))
        .values('month', 'character__character_name')
        .annotate(total=Sum('duration'))
    )
    for row in qs:
        month = row['month'].strftime("%Y-%m")
        result[month][row['character__character_name']] += row['total'].total_seconds()
    return result


class StatsResponse(Schema):
    fleet_seconds_by_hull_by_month: dict[str, dict[str, float]]
    xes_by_hull_by_month: dict[str, dict[str, float]]
    fleet_seconds_by_month: dict[str, float]
    pilots_by_month: dict[str, float]
    xes_by_hull_28d: dict[str, float]
    fleet_seconds_by_hull_28d: dict[str, float]
    x_vs_time_by_hull_28d: dict[str, dict[str, float]]
    time_spent_in_fleet_by_month: dict[str, dict[str, float]]


api = NinjaAPI()


def setup(api: NinjaAPI) -> None:
    StatisticsAPIEndpoints(api)


class StatisticsAPIEndpoints:

    tags = ["Statistics"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/stats", response={200: StatsResponse, 403: dict}, tags=self.tags)
        def statistics(request):
            if not request.user.has_perm("incursions.stats_view"):
                return 403, {"error": "Permission denied"}

            return StatsResponse(
                fleet_seconds_by_hull_by_month=get_fleet_seconds_by_hull_by_month,
                xes_by_hull_by_month=get_xes_by_hull_by_month,
                fleet_seconds_by_month=get_fleet_seconds_by_month,
                pilots_by_month=get_pilots_by_month,
                xes_by_hull_28d=get_xes_by_hull_28d,
                fleet_seconds_by_hull_28d=get_fleet_seconds_by_hull_28d,
                x_vs_time_by_hull_28d=get_x_vs_time_by_hull_28d,
                time_spent_in_fleet_by_month=get_time_spent_in_fleet_by_month
            )



from datetime import datetime, timezone

from ninja import NinjaAPI, Schema

from django.db import transaction

from allianceauth.eveonline.models import EveCharacter
from allianceauth.framework.api.user import get_main_character_from_user
from allianceauth.services.hooks import get_extension_logger

from incursions.api.schema import (
    AllianceSchema, CharacterSchema, CorporationSchema,
)
from incursions.models.waitlist import Ban


class PublicBanSchema(Schema):
    id: int
    entity_name: str
    entity_type: str
    entity_character: CharacterSchema | None
    entity_corporation: CorporationSchema | None
    entity_alliance: AllianceSchema | None
    issued_at: datetime
    issued_by: CharacterSchema | None
    public_reason: str | None
    revoked_at: datetime | None
    revoked_by: CharacterSchema | None


class BanSchema(Schema):
    id: int
    entity_name: str
    entity_type: str
    entity_character: CharacterSchema | None
    entity_corporation: CorporationSchema | None
    entity_alliance: AllianceSchema | None
    issued_at: datetime
    issued_by: CharacterSchema | None
    reason: str
    public_reason: str | None
    revoked_at: datetime | None
    revoked_by: CharacterSchema | None


class UpdateBanSchema(Schema):
    id: int
    entity_type: str
    entity_character: CharacterSchema | None
    entity_corporation: CorporationSchema | None
    entity_alliance: AllianceSchema | None
    reason: str
    public_reason: str | None
    revoked_at: datetime | None


logger = get_extension_logger(__name__)
api = NinjaAPI()


def setup(api: NinjaAPI) -> None:
    BansAPIEndpoints(api)


class BansAPIEndpoints:

    tags = ["Bans"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/bans", response={200: list[BanSchema], 403: dict}, tags=self.tags)
        def list_bans(request):
            if not request.user.has_perm("incursions.view_bans"):
                logger.warning(f"User {request.user} denied listing bans")
                return 403, {"error": "Permission denied"}

            bans = Ban.objects.filter(revoked_at__isnull=True).select_related("entity_character", "entity_corporation", "entity_alliance", "issued_by", "revoked_by")
            logger.info(f"User {request.user} listed {bans.count()} active bans")
            return 200, list(bans)

        @api.post("/bans", response={200: dict, 403: dict}, tags=self.tags)
        def create_ban(request, payload: UpdateBanSchema):
            if not request.user.has_perm("incursions.manage_bans"):
                logger.warning(f"User {request.user} denied ban creation")
                return 403, {"error": "Permission denied"}

            with transaction.atomic():
                Ban.objects.create(
                    pk=payload.id,
                    entity_type=payload.entity_type,
                    issued_at=datetime.now(timezone.utc),
                    issued_by=get_main_character_from_user(request.user),
                    entity_character_id=payload.entity_character.pk if payload.entity_character else None,
                    entity_corporation_id=payload.entity_corporation.pk if payload.entity_corporation else None,
                    entity_alliance_id=payload.entity_alliance.pk if payload.entity_alliance else None,
                    reason=payload.reason,
                    public_reason=payload.public_reason,
                    revoked_at=payload.revoked_at,
                    revoked_by=get_main_character_from_user(request.user) if payload.revoked_at else None
                )
            logger.info(f"Ban {payload.id} created by user {request.user}")
            return 200, {"status": "Ban created"}

        @api.get("/bans/{character_id}", response={200: list[BanSchema], 403: dict, 404: dict}, tags=self.tags)
        def character_history(request, character_id: int):
            if not request.user.has_perm("incursions.view_bans"):
                logger.warning(f"User {request.user} denied access to character ban history for {character_id}")
                return 403, {"error": "Permission denied"}

            if not EveCharacter.objects.filter(character_id=character_id).exists():
                logger.error(f"Character {character_id} not found")
                return 404, {"error": "Character not found"}

            bans = Ban.objects.filter(entity_character__character_id=character_id, entity_type="Character").select_related("entity_character", "entity_corporation", "entity_alliance", "issued_by", "revoked_by")
            logger.info(f"User {request.user} retrieved {bans.count()} bans for character {character_id}")
            return 200, list(bans)

        @api.patch("/bans/{ban_id}", response={200: dict, 403: dict, 404: dict}, tags=self.tags)
        def update_ban(request, ban_id: int, payload: BanSchema):
            if not request.user.has_perm("incursions.manage_bans"):
                logger.warning(f"User {request.user} denied ban update")
                return 403, {"error": "Permission denied"}

            try:
                ban = Ban.objects.select_for_update().get(pk=ban_id)
            except Ban.DoesNotExist:
                logger.error(f"Ban {ban_id} not found")
                return 404, {"error": "Ban not found"}

            ban.reason = payload.reason
            ban.public_reason = payload.public_reason
            ban.revoked_at = payload.revoked_at
            ban.revoked_by = get_main_character_from_user(request.user) if payload.revoked_at else None
            ban.issued_by = get_main_character_from_user(request.user)
            ban.issued_at = datetime.now(timezone.utc)
            ban.save(update_fields=["reason", "public_reason", "revoked_at", "revoked_by", "issued_by", "issued_at"])

            logger.info(f"Ban {ban_id} updated by user {request.user}")
            return 200, {"status": "Ban updated"}

        @api.delete("/bans/{ban_id}", response={200: dict, 403: dict, 404: dict}, tags=self.tags)
        def revoke_ban(request, ban_id: int):
            if not request.user.has_perm("incursions.manage_bans"):
                logger.warning(f"User {request.user} denied ban revocation")
                return 403, {"error": "Permission denied"}

            try:
                ban = Ban.objects.select_for_update().get(pk=ban_id)
            except Ban.DoesNotExist:
                logger.error(f"Ban {ban_id} not found")
                return 404, {"error": "Ban not found"}

            ban.revoked_at = datetime.now(timezone.utc)
            ban.revoked_by = get_main_character_from_user(request.user)
            ban.save(update_fields=["revoked_at", "revoked_by"])

            logger.info(f"Ban {ban_id} revoked by user {request.user}")
            return 200, {"status": "Ban revoked"}



from corptools.models import EveItemType
from ninja import NinjaAPI, Schema

from allianceauth.services.hooks import get_extension_logger

from incursions.api.schema import CharacterSchema, HullSchema
from incursions.api.skills import EVECATEGORY_SKILLS, RELEVANT_SKILL_GROUPS
from incursions.models.waitlist import (
    FittingHistory, FleetActivity, SkillHistory,
)

logger = get_extension_logger(__name__)
api = NinjaAPI()


class ActivityEntrySchema(Schema):
    hull: HullSchema
    logged_at: int
    time_in_fleet: int


class ActivitySummaryEntrySchema(Schema):
    hull: HullSchema
    time_in_fleet: int


class ActivityResponse(Schema):
    activity: list[ActivityEntrySchema]
    summary: list[ActivitySummaryEntrySchema]


class FleetCompEntrySchema(Schema):
    hull: HullSchema
    character: CharacterSchema
    logged_at: int
    time_in_fleet: int
    is_boss: bool


class FleetCompResponse(Schema):
    fleets: dict[int, list[FleetCompEntrySchema]]


class XupHistoryLineSchema(Schema):
    logged_at: int
    dna: str
    implants: list[int]
    hull: HullSchema


class XupHistorySchema(Schema):
    xups: list[XupHistoryLineSchema]


class SkillHistoryResponseLineSchema(Schema):
    skill_id: int
    old_level: int
    new_level: int
    logged_at: int
    name: str


class SkillHistoryResponse(Schema):
    history: list[SkillHistoryResponseLineSchema]
    ids: dict[str, int]


def setup(api: NinjaAPI) -> None:
    HistoryAPIEndpoints(api)


class HistoryAPIEndpoints:

    tags = ["History"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/history/fleet", response=ActivityResponse, tags=self.tags)
        def fleet_history(request, character_id: int):
            if not request.user.has_perm("incursions.waitlist_history_view"):
                logger.warning(f"User {request.user} denied access to fleet history for character {character_id}")
                return 403, {"error": "Permission denied"}

            activities = FleetActivity.objects.filter(character_id=character_id).select_related("ship").order_by("-first_seen")
            time_by_hull: dict[int, int] = {}
            activity_entries: list[ActivityEntrySchema] = []

            for act in activities:
                time_in_fleet = act.last_seen - act.first_seen
                hull_id = act.ship.type_id
                hull_name = act.ship.name
                time_by_hull[hull_id] = time_by_hull.get(hull_id, 0) + time_in_fleet

                activity_entries.append(ActivityEntrySchema(
                    hull=HullSchema(id=hull_id, name=hull_name),
                    logged_at=act.first_seen,
                    time_in_fleet=time_in_fleet,
                ))

            summary_entries: list[ActivitySummaryEntrySchema] = [
                ActivitySummaryEntrySchema(hull=HullSchema(id=hull_id, name=EveItemType.objects.only("name").get(pk=hull_id).name), time_in_fleet=total_time)
                for hull_id, total_time in time_by_hull.items()
            ]
            summary_entries.sort(key=lambda s: s.time_in_fleet, reverse=True)

            logger.info(f"Fleet history returned for character {character_id} with {len(activity_entries)} entries")
            return {"activity": activity_entries, "summary": summary_entries}

        @api.get("/history/fleet-comp", response=FleetCompResponse, tags=self.tags)
        def fleet_comp(request, time: int):
            if not request.user.has_perm("incursions.waitlist_history_view"):
                logger.warning(f"User {request.user} denied access to fleet composition at time {time}")
                return 403, {"error": "Permission denied"}

            comp_entries = FleetActivity.objects.select_related("character", "ship", "fleet", "fleet__boss").filter(first_seen__lte=time, last_seen__gte=time)
            fleets: dict[int, list[FleetCompEntrySchema]] = {}

            for entry in comp_entries:
                fleet_id = entry.fleet.pk
                is_boss = entry.fleet.boss_id == entry.character.pk
                fleets.setdefault(fleet_id, []).append(FleetCompEntrySchema(
                    hull=HullSchema.from_orm(entry.ship),
                    character=CharacterSchema.from_orm(entry.character),
                    logged_at=entry.first_seen,
                    time_in_fleet=entry.last_seen - entry.first_seen,
                    is_boss=is_boss,
                ))

            logger.info(f"Fleet composition at {time} returned with {sum(len(v) for v in fleets.values())} members")
            return {"fleets": fleets}

        @api.get("/history/xup", response=XupHistorySchema, tags=self.tags)
        def xup_history(request, character_id: int):
            if not request.user.has_perm("incursions.waitlist_history_view"):
                logger.warning(f"User {request.user} denied access to XUP history for character {character_id}")
                return 403, {"error": "Permission denied"}

            xup_lines: list[XupHistoryLineSchema] = []
            fittings = FittingHistory.objects.filter(character_id=character_id).select_related("fit__ship", "implant_set").order_by("-pk")

            for xup in fittings:
                try:
                    implants_list = [int(i) for i in xup.implant_set.implants.split(':') if i]
                except Exception:
                    implants_list = []

                hull_schema = HullSchema().from_orm(xup.fit.ship)
                xup_lines.append(XupHistoryLineSchema(logged_at=xup.logged_at, dna=xup.fit.dna, implants=implants_list, hull=hull_schema))

            logger.info(f"XUP history returned for character {character_id} with {len(xup_lines)} entries")
            return {"xups": xup_lines}

        @api.get("/history/skills", response=SkillHistoryResponse, tags=self.tags)
        def skill_history(request, character_id: int):
            if not request.user.has_perm("incursions.waitlist_history_view"):
                logger.warning(f"User {request.user} denied access to skill history for character {character_id}")
                return 403, {"error": "Permission denied"}

            history_list: list[SkillHistoryResponseLineSchema] = []
            skills = SkillHistory.objects.filter(character_id=character_id).select_related("skill__group__category").order_by("-pk")
            for skill in skills:
                if skill.skill.group.category.category_id in RELEVANT_SKILL_GROUPS:
                    history_list.append(SkillHistoryResponseLineSchema(
                        skill_id=skill.skill.type_id,
                        old_level=skill.old_level,
                        new_level=skill.new_level,
                        logged_at=skill.logged_at,
                        name=skill.skill.name,
                    ))

            ids = {
                skill.name: skill.type_id
                for skill in EveItemType.objects.filter(group__category_id=EVECATEGORY_SKILLS).only("name", "type_id")
            }

            logger.info(f"Skill history returned for character {character_id} with {len(history_list)} entries")
            return {"history": history_list, "ids": ids}



from corptools.models import EveItemType
from ninja import NinjaAPI, Schema

from django.db import transaction
from django.http import Http404

from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger

from incursions.api.schema import CharacterSchema, HullSchema
from incursions.models.waitlist import (
    ActiveFleet, Fleet, FleetSquad, WaitlistCategory,
)
from incursions.providers import kick_all_fleet_members


class FleetStatusFleetSchema(Schema):
    id: int
    boss: CharacterSchema


class FleetStatusResponse(Schema):
    fleets: list[FleetStatusFleetSchema]


class FleetInfoSquadSchema(Schema):
    id: int
    name: str


class FleetInfoWingSchema(Schema):
    id: int
    name: str
    squads: list[FleetInfoSquadSchema]


class FleetInfoResponse(Schema):
    fleet_id: int
    wings: list[FleetInfoWingSchema]


class FleetMemberSchema(Schema):
    id: int
    name: str | None = None
    ship: HullSchema
    role: str


class FleetCompSquadMembersSchema(Schema):
    id: int
    name: str
    members: list[FleetMemberSchema]


class FleetCompWingSchema(Schema):
    id: int
    name: str
    squads: list[FleetCompSquadMembersSchema]
    member: FleetMemberSchema | None = None


class FleetCompResponse(Schema):
    wings: list[FleetCompWingSchema]
    id: int
    member: FleetMemberSchema | None = None


class FleetMembersMemberSchema(Schema):
    id: int
    name: str | None = None
    ship: HullSchema
    wl_category: str | None = None
    category: str | None = None
    role: str


class FleetMembersResponse(Schema):
    members: list[FleetMembersMemberSchema]


class RegisterRequest(Schema):
    character_id: int
    fleet_id: int
    assignments: dict[str, tuple[int, int]]


logger = get_extension_logger(__name__)
api = NinjaAPI()


def setup(api: NinjaAPI) -> None:
    FleetAPIEndpoints(api)


class FleetAPIEndpoints:

    tags = ["Fleets"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/status", response={200: FleetStatusResponse, 403: dict}, tags=self.tags)
        def fleet_status(request):
            if not request.user.has_perm("incursions.basic_waitlist"):
                logger.warning(f"User {request.user} denied access to fleet status")
                return 403, {"error": "Permission denied"}

            fleet = ActiveFleet.get_solo().fleet
            logger.info(f"User {request.user} retrieved active fleet {fleet.pk}")
            return FleetStatusResponse(fleets=[fleet])

        @api.get("/info", response={200: FleetInfoResponse, 403: dict}, tags=self.tags)
        def fleet_info(request, boss_character_id: int):
            if not request.user.has_perm("incursions.waitlist_manage_waitlist"):
                logger.warning(f"User {request.user} denied access to fleet info")
                return 403, {"error": "Permission denied"}

            fleet = ActiveFleet.get_solo().fleet
            fleet_wings = fleet.get_fleet_wings()
            wings = [
                FleetInfoWingSchema(
                    id=wing.id,
                    name=wing.name,
                    squads=[FleetInfoSquadSchema(id=s.id, name=s.name) for s in wing.squads],
                )
                for wing in fleet_wings
            ]
            logger.info(f"Fleet info fetched for fleet {fleet.pk} by user {request.user}")
            return FleetInfoResponse(fleet_id=fleet.pk, wings=wings)

        @api.get("/fleetcomp", response=FleetCompResponse, tags=self.tags)
        def fleet_composition(request, boss_character_id: int):
            if not request.user.has_perm("incursions.waitlist_fleet_view"):
                logger.warning(f"User {request.user} denied access to fleet composition")
                return 403, {"error": "Permission denied"}

            fleet = ActiveFleet.get_solo().fleet
            fleet_members = fleet.get_fleet_members()
            fleet_wings = fleet.get_fleet_wings()
            wings: list[FleetCompWingSchema] = []

            for wing in fleet_wings:
                squads: list[FleetMemberSchema] = []
                for fm in fleet_members:
                    if fm.wing_id == wing.id:
                        try:
                            character = EveCharacter.objects.only("character_name").get(pk=fm.character_id)
                            ship = EveItemType.objects.only("name").get(pk=fm.ship_type_id)
                        except (EveCharacter.DoesNotExist, EveItemType.DoesNotExist):
                            continue
                        squads.append(FleetMemberSchema(
                            id=fm.squad_id,
                            name=character.character_name,
                            ship=HullSchema(id=ship.pk, name=ship.name),
                            role=fm.role_name,
                        ))
                wings.append(FleetCompWingSchema(id=wing.id, name=wing.name, squads=squads))

            logger.info(f"Fleet composition fetched for fleet {fleet.pk} by user {request.user}")
            return FleetCompResponse(wings=wings, id=fleet.pk)

        @api.get("/fleet/members/{character_id}", response=FleetMembersResponse, tags=self.tags)
        def fleet_members(request, character_id: int):
            if not request.user.has_perm("incursions.waitlist_fleet_view"):
                logger.warning(f"User {request.user} denied access to fleet members for character {character_id}")
                return 403, {"error": "Permission denied"}

            try:
                fleet = Fleet.objects.select_related("boss").get(boss__character_id=character_id)
            except Fleet.DoesNotExist:
                logger.error(f"Fleet not found for character {character_id}")
                raise Http404("Fleet not found")

            fleet_members = fleet.get_fleet_members()
            wings = {w.id: w.name for w in fleet.get_fleet_wings()}
            members: list[FleetMembersMemberSchema] = []

            for fm in fleet_members:
                try:
                    char = EveCharacter.objects.only("character_name").get(pk=fm.character_id)
                    ship = EveItemType.objects.only("name").get(pk=fm.ship_type_id)
                    category = FleetSquad.objects.only("category").get(squad_id=fm.squad_id)
                except (EveCharacter.DoesNotExist, EveItemType.DoesNotExist, FleetSquad.DoesNotExist):
                    continue

                members.append(FleetMembersMemberSchema(
                    id=fm.character_id,
                    name=char.character_name,
                    ship=HullSchema(id=ship.pk, name=ship.name),
                    wl_category=category.category.name,
                    category=f"{wings.get(fm.wing_id)} - {fm.squad_name}",
                    role=fm.role,
                ))

            logger.info(f"Fleet members listed for fleet {fleet.pk} by user {request.user}")
            return FleetMembersResponse(members=members)

        @api.post("/register", tags=self.tags)
        def register_fleet(request, body: RegisterRequest):
            if not request.user.has_perm("incursions.waitlist_manage_waitlist"):
                logger.warning(f"User {request.user} denied fleet registration")
                return 403, {"error": "Permission denied"}

            with transaction.atomic():
                fleet, _ = Fleet.objects.select_for_update().get_or_create(
                    pk=body.fleet_id,
                    defaults={"boss": EveCharacter.objects.get(pk=body.character_id)},
                )

                active_fleet = ActiveFleet.get_solo()
                active_fleet.fleet = fleet
                active_fleet.save(update_fields=["fleet"])

                for category, (wing_id, squad_id) in body.assignments.items():
                    cat = WaitlistCategory.objects.get(name=category)
                    FleetSquad.objects.update_or_create(
                        fleet=fleet,
                        category=cat,
                        defaults={"wing_id": wing_id, "squad_id": squad_id},
                    )

            logger.info(f"Fleet {fleet.pk} registered and squads assigned by user {request.user}")
            return "OK"

        @api.post("/close", tags=self.tags)
        def close_fleet(request, boss_character_id: int):
            if not request.user.has_perm("incursions.waitlist_manage_waitlist"):
                logger.warning(f"User {request.user} denied closing fleet")
                return 403, {"error": "Permission denied"}

            fleet = ActiveFleet.get_solo().fleet
            count = kick_all_fleet_members(boss_character_id=fleet.boss.character_id, fleet_id=fleet.pk)
            logger.info(f"User {request.user} closed fleet {fleet.pk} and kicked {count} members")
            return f"Kicked {count} Fleet Members"



import json
from datetime import datetime

from corptools.models import EveItemType
from ninja import NinjaAPI, Schema

from django.db import transaction
from django.db.models import Prefetch
from django.http import HttpRequest
from django.shortcuts import get_object_or_404
from django.utils.timezone import now

from allianceauth.eveonline.models import EveCharacter
from allianceauth.framework.api.evecharacter import (
    get_main_character_from_evecharacter,
)
from allianceauth.framework.api.user import (
    get_all_characters_from_user, get_main_character_from_user,
)
from allianceauth.services.hooks import get_extension_logger

from incursions.api.helpers.fittings import FittingParser
from incursions.api.helpers.sse import SSEClient, SSEEvent
from incursions.api.schema import CharacterSchema, HullSchema
from incursions.app_settings import SSE_SECRET, SSE_SITE_URL
from incursions.models.waitlist import (
    ActiveFleet, Fitting, FleetSquad, ImplantSet, Waitlist, WaitlistCategory,
    WaitlistCategoryRule, WaitlistEntry, WaitlistEntryFit,
)
from incursions.providers import get_character_implants, invite_to_fleet

logger = get_extension_logger(__name__)
sse_client = SSEClient(SSE_SITE_URL, SSE_SECRET)


# Schemas
class WaitlistEntryFitSchema(Schema):
    id: int
    approved: bool
    category: str
    dna: str
    hull: HullSchema | None = None
    character: CharacterSchema | None = None
    hours_in_fleet: int | None = None
    review_comment: str | None = None
    implants: list[int] | None = None
    fit_analysis: dict | None = None
    is_alt: bool
    messagexup: str | None = None
    tags: list[str] = []


class WaitlistEntrySchema(Schema):
    id: int
    character: CharacterSchema | None = None
    joined_at: datetime
    can_remove: bool
    fits: list[WaitlistEntryFitSchema]


class WaitlistResponse(Schema):
    open: bool
    waitlist: list[WaitlistEntrySchema] | None = None
    categories: list[str]


class XupRequest(Schema):
    waitlist_id: int
    character_id: int
    eft: str
    is_alt: bool
    messagexup: str


class RemoveXRequest(Schema):
    id: int


class RemoveFitRequest(Schema):
    id: int


class InviteRequest(Schema):
    id: int
    character_id: int


class ApproveRequest(Schema):
    id: int


class RejectRequest(Schema):
    id: int
    review_comment: str


class SetOpenRequest(Schema):
    open: bool
    waitlist_id: int


class EmptyWaitlistRequest(Schema):
    waitlist_id: int


class WaitlistUpdateSchema(Schema):
    waitlist_id: int


api = NinjaAPI()


def setup(api: NinjaAPI) -> None:
    WaitlistAPIEndpoints(api)


class WaitlistAPIEndpoints:
    tags = ["Waitlist"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get(
            "/waitlist",
            response={200: WaitlistResponse, 403: dict, 404: dict},
            tags=self.tags,
        )
        def get_waitlist(request: HttpRequest) -> WaitlistResponse | tuple[int, dict]:
            logger.info("Fetching waitlist for user %s", request.user)
            if not request.user.has_perm("incursions.basic_waitlist"):
                logger.warning("User %s does not have permission incursions.basic_waitlist", request.user)
                return 403, {"error": "Permission denied"}

            waitlist = Waitlist.get_solo()
            categories: list[str] = list(WaitlistCategory.objects.values_list("name", flat=True))
            request_main_character = get_main_character_from_user(request.user)
            context_manager: bool = request.user.has_perm("incursions.waitlist_manage_waitlist")
            context_wl_approver: bool = context_manager or request.user.has_perm("incursions.waitlist_manage_waitlist_approve_fits")
            context_implants: bool = context_wl_approver or request.user.has_perm("incursions.waitlist_implants_view")
            context_stats: bool = context_wl_approver or request.user.has_perm("incursions.waitlist_stats_view")
            context_quantity: bool = context_wl_approver or request.user.has_perm("incursions.waitlist_context_a")
            context_hull: bool = context_wl_approver or request.user.has_perm("incursions.waitlist_context_b")
            context_time: bool = context_wl_approver or request.user.has_perm("incursions.waitlist_context_c")
            context_name: bool = context_wl_approver or request.user.has_perm("incursions.waitlist_context_d")

            if not waitlist.is_open:
                logger.info("Waitlist is closed.")
                return WaitlistResponse(open=False, waitlist=None, categories=categories)

            waitlist_entries = (
                WaitlistEntry.objects.filter(waitlist=waitlist)
                .select_related("main_character")
                .prefetch_related(
                    Prefetch(
                        "waitlistentryfit_set",
                        queryset=WaitlistEntryFit.objects.select_related(
                            "fit", "fit__ship", "character", "category", "implant_set"
                        ).order_by("pk"),
                    )
                )
                .order_by("pk")
            )

            response_waitlist: list[WaitlistEntrySchema] = []
            for waitlist_entry in waitlist_entries:
                fits_data: list[WaitlistEntryFitSchema] = []
                is_self: bool = request_main_character == waitlist_entry.main_character
                for wl_fit in waitlist_entry.waitlistentryfit_set.all():
                    if wl_fit.implant_set and wl_fit.implant_set.implants:
                        implants_list = json.loads(wl_fit.implant_set.implants)
                    else:
                        implants_list: list[str] = []
                    try:
                        fit_analysis_data = json.loads(wl_fit.fit_analysis)
                    except Exception as e:
                        logger.error("Error parsing fit_analysis for fit %s: %s", wl_fit.pk, e)
                        fit_analysis_data = None
                    fits_data.append(
                        WaitlistEntryFitSchema(
                            id=wl_fit.pk,
                            approved=wl_fit.approved,
                            category=wl_fit.category.name,
                            hull=HullSchema(
                                id=wl_fit.fit.ship.type_id
                                if (is_self or context_wl_approver or context_hull)
                                else 0,
                                name=wl_fit.fit.ship.name
                                if (is_self or context_wl_approver or context_hull)
                                else "Hidden",
                            ),
                            character=CharacterSchema(
                                character_name=wl_fit.character.character_name
                                if (is_self or context_name)
                                else "Hidden",
                                character_id=wl_fit.character.character_id
                                if (is_self or context_name)
                                else 0,
                            ),
                            hours_in_fleet=(wl_fit.cached_time_in_fleet // 3600)
                            if context_stats
                            else 0,
                            review_comment=wl_fit.review_comment if (is_self or context_wl_approver) else None,
                            dna=wl_fit.fit.dna if (is_self or context_wl_approver) else "",
                            implants=implants_list if (is_self or context_wl_approver or context_implants) else [],
                            fit_analysis=fit_analysis_data if (is_self or context_wl_approver or context_implants) else None,
                            is_alt=wl_fit.is_alt if (is_self or context_wl_approver or context_quantity) else False,
                            messagexup=wl_fit.messagexup if (is_self or context_wl_approver) else None,
                            tags=[tag for tag in wl_fit.tags.split(",") if tag],
                        )
                    )
                response_waitlist.append(
                    WaitlistEntrySchema(
                        id=waitlist_entry.pk,
                        character=CharacterSchema(
                            character_name=waitlist_entry.main_character.character_name
                            if (is_self or context_name)
                            else "Hidden",
                            character_id=waitlist_entry.main_character.character_id
                            if (is_self or context_name)
                            else 0,
                        ),
                        joined_at=waitlist_entry.joined_at if (is_self or context_time) else now(),
                        can_remove=True if (is_self or context_wl_approver) else False,
                        fits=fits_data,
                    )
                )
            logger.info("Returning waitlist with %d entries", len(response_waitlist))
            return WaitlistResponse(open=True, waitlist=response_waitlist, categories=categories)

        @api.post(
            "/waitlist/xup",
            response={200: dict, 403: dict, 404: dict},
            tags=self.tags,
        )
        def xup(request: HttpRequest, payload: XupRequest) -> tuple[int, dict]:
            logger.info(
                "Processing xup for waitlist_id: %s and character_id: %s",
                payload.waitlist_id,
                payload.character_id,
            )
            if not request.user.has_perm("incursions.basic_waitlist"):
                logger.warning("User %s lacks basic_waitlist permission", request.user)
                return 403, {"error": "Permission denied"}

            waitlist = get_object_or_404(Waitlist, pk=payload.waitlist_id)
            character = get_object_or_404(EveCharacter, character_id=payload.character_id)
            fitting_parser = FittingParser.from_eft(payload.eft)
            with transaction.atomic():
                fitting, _ = Fitting.objects.get_or_create(
                    dna=fitting_parser.to_dna(),
                    ship=EveItemType.objects.get_or_create_from_esi(fitting_parser.hull)[0],
                )
                implant_set, _ = ImplantSet.objects.get_or_create(
                    implants=list(get_character_implants(character.character_id)[0])
                )
                try:
                    rule = (
                        WaitlistCategoryRule.objects.select_related("waitlist_category")
                        .get(ship=fitting.ship)
                    )
                    waitlist_category = rule.waitlist_category
                except WaitlistCategoryRule.DoesNotExist:
                    waitlist_category, _ = WaitlistCategory.objects.get_or_create(name="Other")
                waitlist_entry, _ = WaitlistEntry.objects.get_or_create(
                    main_character=character,
                    waitlist=waitlist,
                    defaults={"joined_at": now()},
                )
                new_fit = WaitlistEntryFit.objects.create(
                    character=character,
                    waitlist=waitlist,
                    fit=fitting,
                    implant_set=implant_set,
                    approved=False,
                    tags="",
                    category=waitlist_category,
                    cached_time_in_fleet=0,
                    is_alt=payload.is_alt,
                    messagexup=payload.messagexup,
                )
            sse_payload = WaitlistUpdateSchema(waitlist_id=waitlist.pk).model_dump()
            sse_client.submit(
                [SSEEvent.new_json(topic="waitlist", event="waitlist_update", data=sse_payload)]
            )
            boss_char = get_main_character_from_evecharacter(ActiveFleet.get_solo().fleet.boss)
            sse_client.submit(
                [
                    SSEEvent.new(
                        topic=f"account;{boss_char.character_id}",
                        event="message",
                        data="New x-up in waitlist",
                    )
                ]
            )
            logger.info("X-up recorded with fit_id %s", new_fit.pk)
            return 200, {"status": "X-up recorded", "fit_id": new_fit.pk}

        @api.post(
            "/waitlist/remove_x",
            response={200: dict, 403: dict},
            tags=self.tags,
        )
        def remove_x(request: HttpRequest, payload: RemoveXRequest) -> tuple[int, dict]:
            logger.info("Removing waitlist entry x with id %s", payload.id)
            waitlist_entry = get_object_or_404(WaitlistEntry, pk=payload.id)
            waitlist = get_object_or_404(Waitlist, pk=waitlist_entry.waitlist.pk)
            if not (
                waitlist_entry.main_character == get_main_character_from_user(request.user)
                or request.user.has_perm("incursions.manage_waitlist")
            ):
                logger.warning("User %s not authorized to remove waitlist entry %s", request.user, payload.id)
                return 403, {"error": "Permission denied"}
            with transaction.atomic():
                waitlist_entry.delete()
            sse_payload = WaitlistUpdateSchema(waitlist_id=waitlist.pk).model_dump()
            sse_client.submit(
                [SSEEvent.new_json(topic="waitlist", event="waitlist_update", data=sse_payload)]
            )
            logger.info("Waitlist entry %s removed", payload.id)
            return 200, {"status": "Waitlist X-Up removed"}

        @api.post(
            "/waitlist/remove_fit",
            response={200: dict, 403: dict},
            tags=self.tags,
        )
        def remove_fit(request: HttpRequest, payload: RemoveFitRequest) -> tuple[int, dict]:
            logger.info("Removing fit with id %s", payload.id)
            entry = get_object_or_404(WaitlistEntryFit, pk=payload.id)
            waitlist_entry = get_object_or_404(WaitlistEntry, pk=entry.waitlist_entry.pk)
            waitlist = get_object_or_404(Waitlist, pk=waitlist_entry.waitlist.pk)
            if not (entry.character in get_all_characters_from_user(request.user) or request.user.has_perm("incursions.manage_waitlist")):
                logger.warning("User %s not authorized to remove fit %s", request.user, payload.id)
                return 403, {"error": "Permission denied"}
            with transaction.atomic():
                entry.delete()
                if not WaitlistEntryFit.objects.filter(waitlist_entry=waitlist_entry).exists():
                    waitlist_entry.delete()
            sse_payload = WaitlistUpdateSchema(waitlist_id=waitlist.pk).model_dump()
            sse_client.submit([SSEEvent.new_json(topic="waitlist", event="waitlist_update", data=sse_payload)])
            logger.info("Waitlist fit %s removed", payload.id)
            return 200, {"status": "Waitlist X-Up Fit removed"}

        @api.post("/waitlist/invite", tags=self.tags)
        def invite(request: HttpRequest, payload: InviteRequest) -> dict:
            logger.info("Inviting character %s for waitlist entry fit id %s", payload.character_id, payload.id)
            if not request.user.has_perm("incursions.manage_waitlist"):
                logger.warning("User %s not authorized to invite", request.user)
                return 403, {"error": "Permission denied"}
            active_fleet = ActiveFleet.get_solo()
            waitlist_entry_fit = get_object_or_404(WaitlistEntryFit, pk=payload.id)
            fleet_squad = get_object_or_404(
                FleetSquad, fleet=active_fleet.fleet, name=waitlist_entry_fit.category.name
            )
            squad_id = fleet_squad.squad_id
            wing_id = fleet_squad.wing_id
            invite_to_fleet(
                boss_character_id=active_fleet.fleet.boss.pk,
                fleet_id=active_fleet.fleet.pk,
                character_id=payload.character_id,
                squad_id=squad_id if squad_id else None,
                wing_id=wing_id if wing_id else None,
                role="squad_member",
            )
            invited_char = get_main_character_from_evecharacter(waitlist_entry_fit.character)
            sse_client.submit([SSEEvent.new(
                topic=f"account;{invited_char.character_id}",
                event="wakeup",
                data=f"{active_fleet.fleet.boss.character_name} has invited your {waitlist_entry_fit.fit.ship.name} to fleet.",
            )])
            logger.info("Character %s invited for waitlist entry %s", payload.character_id, payload.id)
            return {"status": f"Character {payload.character_id} invited to waitlist entry {payload.id}"}

        @api.post("/waitlist/approve", tags=self.tags)
        def approve_fit(request: HttpRequest, payload: ApproveRequest) -> dict:
            logger.info("Approving fit with id %s", payload.id)
            if not (
                request.user.has_perm("incursions.manage_waitlist")
                or request.user.has_perm("incursions.waitlist_manage_waitlist_approve_fits")
            ):
                logger.warning("User %s not authorized to approve fit %s", request.user, payload.id)
                return 403, {"error": "Permission denied"}
            with transaction.atomic():
                fit = get_object_or_404(WaitlistEntryFit, pk=payload.id)
                fit.approved = True
                fit.save(update_fields=["approved"])
            sse_payload = WaitlistUpdateSchema(waitlist_id=fit.waitlist_entry.pk).model_dump()
            sse_client.submit(
                [SSEEvent.new_json(topic="waitlist", event="waitlist_update", data=sse_payload)]
            )
            logger.info("Fit %s approved", payload.id)
            return {"status": f"Fit {fit.pk} approved"}

        @api.post("/waitlist/reject", tags=self.tags)
        def reject_fit(request: HttpRequest, payload: RejectRequest) -> tuple[int, dict]:
            logger.info("Rejecting fit with id %s", payload.id)
            if not (
                request.user.has_perm("incursions.manage_waitlist")
                or request.user.has_perm("incursions.waitlist_manage_waitlist_approve_fits")
            ):
                logger.warning("User %s not authorized to reject fit %s", request.user, payload.id)
                return 403, {"error": "Permission denied"}
            with transaction.atomic():
                fit = get_object_or_404(WaitlistEntryFit, pk=payload.id)
                fit.approved = False
                fit.review_comment = payload.review_comment
                fit.save(update_fields=["approved", "review_comment"])
            sse_payload = WaitlistUpdateSchema(waitlist_id=fit.waitlist_entry.pk).model_dump()
            sse_client.submit(
                [SSEEvent.new_json(topic="waitlist", event="waitlist_update", data=sse_payload)]
            )
            logger.info("Fit %s rejected", payload.id)
            return 200, {"status": f"Fit {fit.pk} rejected"}

        @api.post("/waitlist/set_open", tags=self.tags)
        def set_open(request: HttpRequest, payload: SetOpenRequest) -> dict:
            logger.info("Setting waitlist %s open state to %s", payload.waitlist_id, payload.open)
            if not request.user.has_perm("incursions.manage_waitlist"):
                logger.warning("User %s not authorized to set open state", request.user)
                return 403, {"error": "Permission denied"}
            with transaction.atomic():
                w = get_object_or_404(Waitlist, pk=payload.waitlist_id)
                w.is_open = payload.open
                w.save(update_fields=["is_open"])
            sse_payload = WaitlistUpdateSchema(waitlist_id=w.pk).model_dump()
            sse_client.submit(
                [SSEEvent.new_json(topic="waitlist", event="open", data=sse_payload)]
            )
            sse_client.submit(
                [SSEEvent.new_json(topic="waitlist", event="waitlist_update", data=sse_payload)]
            )
            logger.info("Waitlist %s open state set to %s", payload.waitlist_id, payload.open)
            return 200, {"status": "OK"}

        @api.post("/waitlist/empty", tags=self.tags)
        def empty_waitlist(request: HttpRequest, payload: EmptyWaitlistRequest) -> dict:
            logger.info("Emptying waitlist with id %s", payload.waitlist_id)
            if not request.user.has_perm("incursions.manage_waitlist"):
                logger.warning("User %s not authorized to empty waitlist", request.user)
                return 403, {"error": "Permission denied"}
            with transaction.atomic():
                w = get_object_or_404(Waitlist, pk=payload.waitlist_id)
                WaitlistEntryFit.objects.filter(waitlist=w).delete()
            sse_payload = WaitlistUpdateSchema(waitlist_id=w.pk).model_dump()
            sse_client.submit(
                [SSEEvent.new_json(topic="waitlist", event="waitlist_update", data=sse_payload)]
            )
            logger.info("Waitlist %s emptied", payload.waitlist_id)
            return 200, {"status": "OK"}
